import typer

from kleinkram.auth.auth import TokenFile

endpoint = typer.Typer(
    name="endpoint",
    help="Get Or Set the current endpoint",
    no_args_is_help=True,
    context_settings={"help_option_names": ["-h", "--help"]},
)


@endpoint.command("set")
def set_endpoint(endpoint: str = typer.Argument(None, help="API endpoint to use")):
    """
    Set the current endpoint

    Use this command to switch between different API endpoints.\n
    Standard endpoints are:\n
    - http://localhost:3000\n
    - https://api.datasets.leggedrobotics.com\n
    - https://api.datasets.dev.leggedrobotics.com
    """

    if not endpoint:
        raise ValueError("No endpoint provided.")

    tokenfile = TokenFile()
    tokenfile.endpoint = endpoint
    tokenfile.writeToFile()

    print()
    print("Endpoint set to: " + endpoint)
    if tokenfile.endpoint not in tokenfile.tokens:
        print("No tokens found for this endpoint.")


@endpoint.command("get")
def get_endpoints():
    """
    Get the current endpoint

    Also displays all endpoints with saved tokens.
    """
    tokenfile = TokenFile()
    print("Current: " + tokenfile.endpoint)
    print()

    if not tokenfile.tokens:
        print("No saved tokens found.")
        return

    print("Saved Tokens found for:")
    for _endpoint, _ in tokenfile.tokens.items():
        print("- " + _endpoint)



import os
from typing import Optional, Annotated

import httpx
import requests
import typer

from kleinkram.api_client import AuthenticatedClient
from kleinkram.error_handling import AccessDeniedException

file = typer.Typer(
    name="file",
    help="File operations",
    no_args_is_help=True,
    context_settings={"help_option_names": ["-h", "--help"]},
)


@file.command("list")
def list_files(
    project: Optional[str] = typer.Option(None, help="Name of Project"),
    mission: Optional[str] = typer.Option(None, help="Name of Mission"),
    topics: Optional[str] = typer.Option(None, help="Comma separated list of topics"),
    tags: Optional[str] = typer.Option(
        None, help="Comma separated list of tagtype:tagvalue pairs"
    ),
):
    """
    List all files with optional filters for project, mission, or topics.

    Can list files of a project, mission, or with specific topics (Logical AND).
    Examples:\n
        - 'klein filelist'\n
        - 'klein file list --project "Project 1"'\n
        - 'klein file list --mission "Mission 1"'\n
        - 'klein file list --topics "/elevation_mapping/semantic_map,/elevation_mapping/elevation_map_raw"'\n
        - 'klein file list --topics "/elevation_mapping/semantic_map,/elevation_mapping/elevation_map_raw" --mission "Mission A"'
    """
    try:
        url = f"/file/filteredByNames"
        params = {}
        if project:
            params["projectName"] = project
        if mission:
            params["missionName"] = mission
        if topics:
            params["topics"] = topics
        if tags:
            params["tags"] = {}
            for tag in tags.split(","):
                tagtype, tagvalue = tag.split("§")
                params["tags"][tagtype] = tagvalue

        client = AuthenticatedClient()
        response = client.get(
            url,
            params=params,
        )
        response.raise_for_status()
        data = response.json()
        missions_by_project_uuid = {}
        files_by_mission_uuid = {}
        for file in data:
            mission_uuid = file["mission"]["uuid"]
            project_uuid = file["mission"]["project"]["uuid"]
            if project_uuid not in missions_by_project_uuid:
                missions_by_project_uuid[project_uuid] = []
            if mission_uuid not in missions_by_project_uuid[project_uuid]:
                missions_by_project_uuid[project_uuid].append(mission_uuid)
            if mission_uuid not in files_by_mission_uuid:
                files_by_mission_uuid[mission_uuid] = []
            files_by_mission_uuid[mission_uuid].append(file)

        print("Files by mission & Project:")
        for project_uuid, missions in missions_by_project_uuid.items():
            first_file = files_by_mission_uuid[missions[0]][0]
            print(f"* {first_file['mission']['project']['name']}")
            for mission in missions:
                print(f"  - {files_by_mission_uuid[mission][0]['mission']['name']}")
                for file in files_by_mission_uuid[mission]:
                    print(f"    - '{file['filename']}'")

    except httpx.HTTPError as e:
        print(f"Failed to fetch missions: {e}")
        raise e



from typing import Annotated

import typer
from rich.table import Table

from kleinkram.api_client import AuthenticatedClient

user = typer.Typer(
    name="users",
    help="User operations",
)


@user.command("list")
def users(ctx: typer.Context):
    """List all users"""

    client = AuthenticatedClient()
    response = client.get("/user/all")
    response.raise_for_status()
    data = response.json()
    table = Table("Name", "Email", "Role", "googleId")
    for user in data:
        table.add_row(user["name"], user["email"], user["role"], user["googleId"])
    print(table)


@user.command("info")
def user_info():
    """Get logged in user info"""
    client = AuthenticatedClient()
    response = client.get("/user/me")
    response.raise_for_status()
    data = response.json()
    print(data)


@user.command("promote")
def promote(email: Annotated[str, typer.Option()]):
    """Promote another user to admin"""
    client = AuthenticatedClient()
    response = client.post("/user/promote", json={"email": email})
    response.raise_for_status()
    print("User promoted.")


@user.command("demote")
def demote(email: Annotated[str, typer.Option()]):
    """Demote another user from admin"""
    client = AuthenticatedClient()
    response = client.post("/user/demote", json={"email": email})
    response.raise_for_status()
    print("User demoted.")



import typing

import typer
from httpx import HTTPStatusError, ReadError, RemoteProtocolError
from rich.console import Console
from rich.panel import Panel
from typer import Typer

from kleinkram.api_client import NotAuthenticatedException

ExceptionType = "typing.Type[Exception]"
ErrorHandlingCallback = typing.Callable[[Exception], int]


class AccessDeniedException(Exception):

    def __init__(self, message: str, api_error: str):
        self.message = message
        self.api_error = api_error


def not_yet_implemented_handler(e: Exception):
    console = Console()
    panel = Panel(
        "This feature is not yet implemented, please check for updates.",
        title="Not Yet Implemented",
        style="yellow",
        padding=(1, 2),
        highlight=True,
    )
    print()
    console.print(panel)
    print()


def not_authenticated_handler(e: NotAuthenticatedException):
    console = Console()
    panel = Panel(
        f"{e.message}\n » Please run 'klein login' to authenticate.",
        title="Not Authenticated",
        style="yellow",
        padding=(1, 2),
        highlight=True,
    )
    print()
    console.print(panel)
    print()


def access_denied_handler(e: AccessDeniedException):
    console = Console()
    panel = Panel(
        f"{e.message}\n » API Response: {e.api_error}",
        title="Access Denied",
        style="red",
        padding=(1, 2),
        highlight=True,
    )
    print()
    console.print(panel)
    print()


def value_error_handler(e: Exception):
    console = Console()
    panel = Panel(
        str(e),
        title="Invalid Argument",
        style="red",
        padding=(1, 2),
        highlight=True,
    )
    print()
    console.print(panel)
    print()


def http_status_error_handler(e: HTTPStatusError):
    console = Console()
    panel = Panel(
        f"An HTTP error occurred: {e}\n\n » Please report this error to the developers.",
        title="HTTP Status Error",
        style="red",
        padding=(1, 2),
        highlight=True,
    )
    print()
    console.print(panel)
    print()


def remote_down_handler(e: Exception):
    console = Console()
    panel = Panel(
        f"An error occurred while communicating with the remote server: {e}\n"
        f"\n » The server may be down or unreachable; please try again.",
        title="Remote Protocol Error",
        style="yellow",
        padding=(1, 2),
        highlight=True,
    )
    print()
    console.print(panel)
    print()


class ErrorHandledTyper(Typer):
    error_handlers: typing.Dict[ExceptionType, ErrorHandlingCallback] = {
        NotAuthenticatedException: not_authenticated_handler,
        AccessDeniedException: access_denied_handler,
        HTTPStatusError: http_status_error_handler,
        NotImplementedError: not_yet_implemented_handler,
        ValueError: value_error_handler,
        RemoteProtocolError: remote_down_handler,
        ReadError: remote_down_handler,
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def error_handler(self, exc: ExceptionType):
        def decorator(f: ErrorHandlingCallback):
            self.error_handlers[exc] = f
            return f

        return decorator

    def __call__(self, *args, **kwargs):
        try:
            super(ErrorHandledTyper, self).__call__(*args, **kwargs)

        except Exception as e:

            # exit with error code 1 if no error handler is defined
            if type(e) not in self.error_handlers:
                typer.secho(
                    f"An unhanded error of type {type(e).__name__} occurred.",
                    fg=typer.colors.RED,
                )
                typer.secho(str(e), fg=typer.colors.RED)
                typer.secho(
                    " » Please report this error to the developers.",
                    fg=typer.colors.RED,
                )
            else:
                self.error_handlers[type(e)](e)

            # exit with error code 1
            exit(1)



from typing import Annotated

import httpx
import typer

from kleinkram.api_client import AuthenticatedClient

project = typer.Typer(
    name="project",
    help="Project operations",
    no_args_is_help=True,
    context_settings={"help_option_names": ["-h", "--help"]},
)


@project.command("list")
def list_projects():
    """
    List all projects.
    """
    try:
        client = AuthenticatedClient()
        response = client.get("/project")
        response.raise_for_status()
        projects = response.json()[0]
        print("Projects:")
        for _project in projects:
            print(f"- {_project['name']}")

    except httpx.HTTPError as e:
        print(f"Failed to fetch projects: {e}")


@project.command("create")
def create_project(
    name: Annotated[str, typer.Option(help="Name of Project")],
    description: Annotated[str, typer.Option(help="Description of Project")],
):
    """
    Create a new project
    """
    # Todo add required tags as option.
    try:
        url = "/project/create"
        client = AuthenticatedClient()
        response = client.post(
            url, json={"name": name, "description": description, "requiredTags": []}
        )  # TODO: Add required tags as option
        if response.status_code >= 400:
            response_json = response.json()
            response_text = response_json["message"]
            print(f"Failed to create project: {response_text}")
            return
        print("Project created")

    except httpx.HTTPError as e:
        print(f"Failed to create project: {e}")
        raise e



import os
from typing import Annotated, Optional

import httpx
import requests
import typer
from rich.console import Console
from rich.table import Table

from kleinkram.api_client import AuthenticatedClient
from kleinkram.error_handling import AccessDeniedException

mission = typer.Typer(
    name="mission",
    help="Mission operations",
    no_args_is_help=True,
    context_settings={"help_option_names": ["-h", "--help"]},
)


@mission.command("tag")
def addTag(
    mission_uuid: Annotated[str, typer.Argument()],
    tagtype_uuid: Annotated[str, typer.Argument()],
    value: Annotated[str, typer.Argument()],
):
    """Tag a mission"""
    try:
        client = AuthenticatedClient()
        response = client.post(
            "/tag/addTag",
            json={"mission": mission_uuid, "tagType": tagtype_uuid, "value": value},
        )
        if response.status_code < 400:
            print("Tagged mission")
        else:
            print(response.json())
            print("Failed to tag mission")
            raise Exception("Failed to tag mission")
    except httpx.HTTPError as e:
        print(e)
        print("Failed to tag mission")
        raise e


@mission.command("list")
def list_missions(
    project: Optional[str] = typer.Option(None, help="Name of Project"),
    verbose: Optional[bool] = typer.Option(
        False, help="Outputs a table with more information"
    ),
):
    """
    List all missions with optional filter for project.
    """

    url = "/mission"
    if project:
        url += f"/filteredByProjectName/{project}"
    else:
        url += "/all"

    client = AuthenticatedClient()

    try:

        response = client.get(url)
        response.raise_for_status()

    except httpx.HTTPError:

        raise AccessDeniedException(
            f"Failed to fetch mission."
            f"Consider using the following command to list all missions: 'klein mission list --verbose'\n",
            f"{response.json()['message']} ({response.status_code})",
        )

    data = response.json()
    missions_by_project_uuid = {}
    for mission in data:
        project_uuid = mission["project"]["uuid"]
        if project_uuid not in missions_by_project_uuid:
            missions_by_project_uuid[project_uuid] = []
        missions_by_project_uuid[project_uuid].append(mission)

    if len(missions_by_project_uuid.items()) == 0:
        print(f"No missions found for project '{project}'. Does it exist?")
        return

    print("missions by Project:")
    if not verbose:
        for project_uuid, missions in missions_by_project_uuid.items():
            print(f"* {missions_by_project_uuid[project_uuid][0]['project']['name']}")
            for mission in missions:
                print(f"  - {mission['name']}")

    else:
        table = Table("UUID", "name", "project", "creator", "createdAt")
        for project_uuid, missions in missions_by_project_uuid.items():
            for mission in missions:
                table.add_row(
                    mission["uuid"],
                    mission["name"],
                    mission["project"]["name"],
                    mission["creator"]["name"],
                    mission["createdAt"],
                )
        console = Console()
        console.print(table)


@mission.command("byUUID")
def mission_by_uuid(
    uuid: Annotated[str, typer.Argument()],
    json: Optional[bool] = typer.Option(False, help="Output as JSON"),
):
    """
    Get mission name, project name, creator and table of its files given a Mission UUID

    Use the JSON flag to output the full JSON response instead.

    Can be run with API Key or with login.
    """
    url = "/mission/byUUID"
    client = AuthenticatedClient()
    response = client.get(url, params={"uuid": uuid})

    try:
        response.raise_for_status()
    except httpx.HTTPError:
        raise AccessDeniedException(
            f"Failed to fetch mission."
            f"Consider using the following command to list all missions: 'klein mission list --verbose'\n",
            f"{response.json()['message']} ({response.status_code})",
        )

    data = response.json()

    if json:
        print(data)
    else:
        print(f"mission: {data['name']}")
        print(f"Creator: {data['creator']['name']}")
        print("Project: " + data["project"]["name"])
        table = Table("Filename", "Size", "date")
        for file in data["files"]:
            table.add_row(file["filename"], f"{file['size']}", file["date"])
        console = Console()
        console.print(table)


@mission.command("download")
def download(
    mission_uuid: Annotated[str, typer.Argument()],
    local_path: Annotated[str, typer.Argument()],
):
    """

    Downloads all files of a mission to a local path.
    The local path must be an empty directory.

    """

    if not os.path.isdir(local_path):
        raise ValueError(f"Local path '{local_path}' is not a directory.")
    if not os.listdir(local_path) == []:
        raise ValueError(f"Local path '{local_path}' is not empty, but must be empty.")

    client = AuthenticatedClient()
    response = client.get("/file/downloadWithToken", params={"uuid": mission_uuid})

    try:
        response.raise_for_status()
    except httpx.HTTPError as e:
        raise AccessDeniedException(
            f"Failed to download file."
            f"Consider using the following command to list all missions: 'klein mission list --verbose'\n",
            f"{response.json()['message']} ({response.status_code})",
        )

    paths = response.json()

    print(f"Downloading files to {local_path}:")
    for path in paths:

        filename = path.split("/")[-1].split("?")[0]
        print(f" - {filename}")

        response = requests.get(path)
        with open(os.path.join(local_path, filename), "wb") as f:
            f.write(response.content)
            print(f"   Downloaded {filename}")
